MODULE stat_utilities
!***********************************************************************
!   FORTRAN 90 module containing commonly used statistical computations.
!
!   Created: 03/01/03 JLH
!
!   LAPACK subroutines xpotrf, xgetrf, xgetri are used in some 
!   subroutines.  These subroutines are available for free download
!   from http://www.netlib.org/lapack/.
!
!   BLAS subroutines xgemv and xgemm are used for matrix - vector and
!   matrix-matrix multiplication.  These are available for free 
!   download from http://www.netlib.org/blas/ .
!
!   In the names of the subroutines above, the x should be replaced
!   with a "s" for single precision or a "d" for double precision.
!
!   Contents:
!        Suboutines for generating random numbers:
!            gen_uniform, gen_normal
!        Smoothing functions:
!             logistic_sm, exponential_sm, quad_log_sm, quartic_sm,
!             Epanechnikov, cubic_sm
!        Subroutines for linear regression:
!             ls_regr_r, ls_regr_d, ls_regr_rmv, ls_regr_dmv
!***********************************************************************
  USE define_types
  USE f90_utilities
!
  IMPLICIT NONE
!***********************************************************************
!  Uniform(a,b) deviates:
!***********************************************************************
  INTERFACE gen_uniform
     MODULE PROCEDURE gen_uniform_r, gen_uniform_d
  END INTERFACE ! rlc gen_uniform
!***********************************************************************
!  Normal(mean, variance) deviates:
!***********************************************************************
  INTERFACE gen_normal
     MODULE PROCEDURE gen_normal_r, gen_normal_d, gen_mvnormal_r,   &
          gen_mvnormal_d
  END INTERFACE ! rlc gen_normal
!***********************************************************************
!  Smoothing functions:
!***********************************************************************
!     Logistic:
!***********************************************************************
  INTERFACE logistic_sm
     MODULE PROCEDURE logistic_sm_r, logistic_sm_d
  END INTERFACE ! rlc logistic_sm
!***********************************************************************
!     Exponential:
!***********************************************************************
  INTERFACE exponential_sm
     MODULE PROCEDURE exponential_sm_r, exponential_sm_d
  END INTERFACE ! rlc exponential_sm
!***********************************************************************
!     Quadratic logarithm:
!***********************************************************************
  INTERFACE quad_log_sm
     MODULE PROCEDURE quad_log_sm_r, quad_log_sm_d
  END INTERFACE ! rlc quad_log_sm
!***********************************************************************
!     Quartic:
!***********************************************************************
  INTERFACE quartic_sm
     MODULE PROCEDURE quartic_sm_r, quartic_sm_d
  END INTERFACE ! rlc quartic_sm
!***********************************************************************
!     Epanechnikov:
!***********************************************************************
  INTERFACE Epanechnikov
     MODULE PROCEDURE Epanechnikov_r, Epanechnikov_d
  END INTERFACE ! rlc Epanechnikov
!***********************************************************************
!      Cubic:
!***********************************************************************
  INTERFACE cubic_sm
     MODULE PROCEDURE cubic_sm_r, cubic_sm_d
  END INTERFACE ! rlc cubic_sm
!***********************************************************************
!   Linear regression:
!***********************************************************************
  INTERFACE ls_regression
     MODULE PROCEDURE ls_regr_r, ls_regr_d, ls_regr_rmv, ls_regr_dmv
  END INTERFACE ! rlc ls_regression
!
CONTAINS
!
  SUBROUTINE gen_uniform_r(a,b,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n uniform(a,b) random variates.
!***********************************************************************
    USE define_types
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: n
    REAL(sp), INTENT(in) :: a, b
    REAL(sp), INTENT(out), DIMENSION(n) :: x
!
    INTEGER(i4b) :: i
    REAL(sp) :: c
!
    DO i = 1,n
       CALL RANDOM_NUMBER(c)
       x(i) = (b - a)*c + a
    END DO
!
    RETURN
  END SUBROUTINE gen_uniform_r
!
  SUBROUTINE gen_uniform_d(a,b,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n uniform(a,b) random variates.
!***********************************************************************
    USE define_types
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: n
    REAL(dp), INTENT(in) :: a, b
    REAL(dp), INTENT(out), DIMENSION(n) :: x
!
    INTEGER(i4b) :: i
    REAL(dp) :: c
!
    DO i = 1,n
       CALL RANDOM_NUMBER(c)
       x(i) = (b - a)*c + a
    END DO
!
    RETURN
  END SUBROUTINE gen_uniform_d
!
  SUBROUTINE gen_normal_r(mean,variance,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n N(mean,variance)
!  random variates.
!***********************************************************************
    USE define_types
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: n
    REAL(sp), INTENT(in) :: mean, variance
    REAL(sp), INTENT(out), DIMENSION(n) :: x
!
    INTEGER(i4b) :: nn, nnb2, i, i1, i2
    REAL(sp) :: s, y
    REAL(sp), DIMENSION(2) :: r
!  
    nn = n - 1
    IF(2*(n/2) .EQ. n) nn = n
    IF(n .EQ. nn) go to 20
!
5   CALL gen_uniform(0.0_sp,1.0_sp,2,r)
    !
    r = 2.0_sp*(r - 0.50_sp)
    s = DOT_PRODUCT(r,r)
    !
    IF(s .GT. 1.0_sp) go to 5
    y = SQRT(-2.0_sp*(LOG(s)/s))
    !
    x(n) = r(1)*y
    IF(n .EQ. 1) go to 99
    !
20  CALL gen_uniform(0.0_sp,1.0_sp,nn,x)
    nnb2 = nn/2
    !
    DO i = 1, nnb2
       i2 = 2*i
       i1 = i2 - 1
12     r(1) = 2.0_sp * (x(i1) - 0.50_sp)
       r(2) = 2.0_sp * (x(i2) - 0.50_sp)
       s = DOT_PRODUCT(r,r)
       IF(s .LE. 1.0_sp) go to 15
       CALL gen_uniform(0.0_sp,1.0_sp,2,x(i1:i2))
       go to 12
15     y = SQRT(-2.0_sp * (LOG(s)/s))
       x(i1) = y*r(1)
       x(i2) = y*r(2)
    END DO
    x = SQRT(variance)*x + mean
!
99  CONTINUE
    RETURN
!
  END SUBROUTINE gen_normal_r
!
  SUBROUTINE gen_normal_d(mean,variance,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n N(mean,variance)
!  random variates.
!***********************************************************************
    USE define_types
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: n
    REAL(dp), INTENT(in) :: mean, variance
    REAL(dp), INTENT(out), DIMENSION(n) :: x
!
    INTEGER(i4b) :: nn, nnb2, i, i1, i2
    REAL(dp) :: s, y
    REAL(dp), DIMENSION(2) :: r
!  
    nn = n - 1
    IF(2*(n/2) .EQ. n) nn = n
    IF(n .EQ. nn) go to 20
!
5   CALL gen_uniform(0.0_dp,1.0_dp,2,r)
!
    r = 2.0_dp*(r - 0.50_dp)
    s = DOT_PRODUCT(r,r)
!
    IF(s .GT. 1.0_dp) go to 5
    y = SQRT(-2.0_dp*(LOG(s)/s))
    !
    x(n) = r(1)*y
    IF(n .EQ. 1) go to 99
!
20  CALL gen_uniform(0.0_dp,1.0_dp,nn,x)
    nnb2 = nn/2
!
    DO i = 1, nnb2
       i2 = 2*i
       i1 = i2 - 1
12     r(1) = 2.0_dp * (x(i1) - 0.50_dp)
       r(2) = 2.0_dp * (x(i2) - 0.50_dp)
       s = DOT_PRODUCT(r,r)
       IF(s .LE. 1.0_dp) go to 15
       CALL gen_uniform(0.0_dp,1.0_dp,2,x(i1:i2))
       go to 12
15     y = SQRT(-2.0_dp * (LOG(s)/s))
       x(i1) = y*r(1)
       x(i2) = y*r(2)
    END DO
    x = SQRT(variance)*x + mean
!
99  CONTINUE
    RETURN
!
  END SUBROUTINE gen_normal_d
!**********************************************************************
!  Multivariate normal:
!***********************************************************************
!  FORTRAN 90 subroutine to generate random normal observations having
!  mean vector mu and covariance matrix cov.  If cov.fac.opt = 0, then
!  the matrix cov contains the Cholesky decomposition of the covariance
!  matrix.  Otherwise, the Cholesky decomposition will be found.
!
!  Input: mean: a k-vector containing the means
!         cov: a k x k matrix containing the covariance matrix if
!              cov_fac_option != 0, and containing the Cholesky
!              decomposition of the covariance matrix otherwise.
!         cov_fac_option: an integer indicating whether or not the
!               matrix cov is the Cholesky decomposition of the 
!               covariance matrix (if 0 then no; otherwise, yes)
!
!  Output: x: an n x k matrix of multivariate normal deviates.
!
!  Algorigthm: Generate k iid E ~ N(0,1)
!              cov = Cholesky decomposition of covariance matrix
!              x( ,1:k) = trans(cov)E + mean ~ MVN(mean,covariance)
!
!  NOTES: uses LAPACK subroutine spotrf (or dpotrf) to compute the 
!         Cholesky decomposition of the covariance matrix.  See
!         http://www.netlib/org/lapack/.
!***********************************************************************
  SUBROUTINE gen_mvnormal_r(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
    USE define_types
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_n, max_k, n, k, cov_fac_option
    REAL(sp), INTENT(in), DIMENSION(k) :: mean
    REAL(sp), INTENT(in), DIMENSION(max_k, k) :: cov
    REAL(sp), INTENT(out), DIMENSION(max_n, k) :: x
!
    INTEGER(i4b) :: info, i, j
    REAL(sp), DIMENSION(:), ALLOCATABLE :: temp, temp_out
    REAL(sp), DIMENSION(:,:), ALLOCATABLE :: cov_fac
!
    EXTERNAL spotrf
!
    ALLOCATE(cov_fac(k,k))
    cov_fac = cov
    DO i = 1, k
    DO j = 1, i-1
       cov_fac(i,j) = 0.0_sp
    END DO
    END DO
!
    IF(cov_fac_option .NE. 0) CALL spotrf("U",k,cov_fac,k,info)
!
    ALLOCATE(temp(k))
    ALLOCATE(temp_out(k))
    DO i = 1,n
       CALL gen_normal(0.0_sp,1.0_sp,k,temp)
       temp = MATMUL(TRANSPOSE(cov_fac),temp)
       x(i,1:k) = temp_out + mean(1:k)
    END DO
    DEALLOCATE(temp, temp_out, cov_fac)
!
!print *, "MATMUL in stat_utilities.f90/gen_mvnormal_r(): OK"
    RETURN
  END SUBROUTINE gen_mvnormal_r
!
  SUBROUTINE gen_mvnormal_d(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
    USE define_types
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_n, max_k, n, k, cov_fac_option
    REAL(dp), INTENT(in), DIMENSION(k) :: mean
    REAL(dp), INTENT(in), DIMENSION(max_k, k) :: cov
    REAL(dp), INTENT(inout), DIMENSION(max_n, k) :: x
!
    INTEGER(i4b) :: info, i, j
    REAL(dp), DIMENSION(:), ALLOCATABLE :: temp, temp_out
    REAL(dp), DIMENSION(:,:), ALLOCATABLE :: cov_fac
!
    EXTERNAL dpotrf
!
    ALLOCATE(cov_fac(k,k))
    cov_fac = cov(1:k,1:k)
    DO i = 1, k
    DO j = 1, i-1
       cov_fac(i,j) = 0.0_dp
    END DO
    END DO
!
    IF(cov_fac_option .NE. 0) CALL dpotrf("U",k,cov_fac,k,info)
!
    ALLOCATE(temp(k))
    ALLOCATE(temp_out(k))
    DO i = 1, n
       CALL gen_normal(0.0_dp,1.0_dp,k,temp)
       temp = MATMUL(TRANSPOSE(cov_fac),temp) 
       x(i,1:k) = temp + mean
    END DO
    DEALLOCATE(temp, temp_out, cov_fac)
!print *, "MATMUL in stat_utilities.f90/gen_mvnormal_d(): OK"
!
    RETURN
  END SUBROUTINE gen_mvnormal_d
!**********************************************************************
!  Smoothing functions:
!**********************************************************************
!    Logisitic:
!**********************************************************************
  FUNCTION logistic_sm_r(x, sm_constant, sm_PARAMETER)
    USE define_types
!
    IMPLICIT NONE
    REAL(sp), INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL(sp) :: logistic_sm_r
!
    logistic_sm_r = 1.0_sp/(1.0_sp + EXP(-sm_PARAMETER*(x - sm_constant)))
!
    RETURN
  END FUNCTION logistic_sm_r
!
  FUNCTION logistic_sm_d(x, sm_constant, sm_PARAMETER)
    USE define_types
!
    IMPLICIT NONE
    REAL(dp), INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL(dp) :: logistic_sm_d
!
    logistic_sm_d = 1.0_dp/(1.0_dp + EXP(-sm_PARAMETER*(x - sm_constant)))
!
    RETURN
  END FUNCTION logistic_sm_d
!**********************************************************************
!    Exponential:
!**********************************************************************
  FUNCTION exponential_sm_r(x, sm_constant, sm_PARAMETER)
    USE define_types
!
    IMPLICIT NONE
    REAL(sp), INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL(sp) :: exponential_sm_r
!
    exponential_sm_r = 1.0_sp - EXP(-sm_PARAMETER*(x - sm_constant)*(x - sm_constant))
!
    RETURN
  END FUNCTION exponential_sm_r
!
  FUNCTION exponential_sm_d(x, sm_constant, sm_PARAMETER)
    USE define_types
!
    IMPLICIT NONE
    REAL(dp), INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL(dp) :: exponential_sm_d
!
    exponential_sm_d = 1.0_dp - EXP(-sm_PARAMETER*(x - sm_constant)*(x - sm_constant))
!
    RETURN
  END FUNCTION exponential_sm_d
!**********************************************************************
!    Quadratic logarithm:
!**********************************************************************
  FUNCTION quad_log_sm_r(x, sm_constant, sm_PARAMETER)
    USE define_types
!
    IMPLICIT NONE
    REAL(sp), INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL(sp) :: quad_log_sm_r
!
    quad_log_sm_r = 1.0_sp/(1.0_sp + EXP(sm_PARAMETER*(x*x - sm_constant*sm_constant)))
!
    RETURN
  END FUNCTION quad_log_sm_r
!
  FUNCTION quad_log_sm_d(x, sm_constant, sm_PARAMETER)
    USE define_types
!
    IMPLICIT NONE
    REAL(dp), INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL(dp):: quad_log_sm_d
!
    quad_log_sm_d = 1.0_dp/(1.0_dp + EXP(sm_PARAMETER*(x*x - sm_constant*sm_constant)))
!
    RETURN
  END FUNCTION quad_log_sm_d
!***********************************************************************
!    Cubic:
!***********************************************************************
  FUNCTION cubic_sm_r(x)
    USE define_types
!
    IMPLICIT NONE
    REAL(sp), INTENT(in) :: x
    REAL(sp) :: cubic_sm_r
!
    IF(ABS(x) .LE. 1.0_sp) THEN
       cubic_sm_r = 0.75_sp*(1.0_sp - x*x)
    ELSE
       cubic_sm_r = 0.0_sp
    END IF
!
    RETURN
  END FUNCTION cubic_sm_r
!
  FUNCTION cubic_sm_d(x)
    USE define_types
!
    IMPLICIT NONE
    REAL(dp), INTENT(in) :: x
    REAL(dp) :: cubic_sm_d
!
    IF(ABS(x) .LE. 1.0_dp) THEN
       cubic_sm_d = 0.75_dp*(1.0_dp - x*x)
    ELSE
       cubic_sm_d = 0.0_dp
    END IF
!
    RETURN
  END FUNCTION cubic_sm_d
!************************************************************************
!     Epanechnikov:
!***********************************************************************
  FUNCTION Epanechnikov_r(x)
    USE define_types
!
    IMPLICIT NONE
    REAL(sp), INTENT(in) :: x
    REAL(sp) :: Epanechnikov_r
!
    IF((x*x) .LE. 0.5_sp) THEN
       Epanechnikov_r = 3.0_sp * (1.0_sp - x**(2.0_sp/5.0_sp))
       Epanechnikov_r = Epanechnikov_r / (4.0_sp*SQRT(5.0_sp))
    ELSE
       Epanechnikov_r = 0.0_sp
    END IF
!
    RETURN
  END FUNCTION Epanechnikov_r
!
  FUNCTION Epanechnikov_d(x)
    USE define_types
!
    IMPLICIT NONE
    REAL(dp), INTENT(in) :: x
    REAL(dp) :: Epanechnikov_d
!
    IF((x*x) .LE. 0.5_dp) THEN
       Epanechnikov_d = 3.0_dp * (1.0_dp - x**(2.0_dp/5.0_dp))
       Epanechnikov_d = Epanechnikov_d / (4.0_dp*SQRT(5.0_dp))
    ELSE
       Epanechnikov_d = 0.0_dp
    END IF
!
    RETURN
  END FUNCTION Epanechnikov_d
!***********************************************************************
!      Quartic:
!***********************************************************************
  FUNCTION quartic_sm_r(x)
    USE define_types
!
    IMPLICIT NONE
    REAL(sp), INTENT(in) :: x
    REAL(sp) :: quartic_sm_r
!
    IF(ABS(x) .LE. 1.0_sp) THEN
       quartic_sm_r = 0.9375_sp*(1.0_sp - x*x)*(1.0_sp - x*x)
    ELSE
       quartic_sm_r = 0.0_sp
    END IF
!
    RETURN
  END FUNCTION quartic_sm_r
!
  FUNCTION quartic_sm_d(x)
    USE define_types
!
    IMPLICIT NONE
    REAL(dp), INTENT(in) :: x
    REAL(dp) :: quartic_sm_d
!
    IF(ABS(x) .LE. 1.0_dp) THEN
       quartic_sm_d = 0.9375_dp*(1.0_dp - x*x)*(1.0_dp - x*x)
    ELSE
       quartic_sm_d = 0.0_dp
    END IF
!
    RETURN
  END FUNCTION quartic_sm_d
!***********************************************************************
!  Linear regression:
!***********************************************************************
!  Notes: Uses LAPACK subroutines sgetrf and sgetri for computing
!         LU factorization and matrix inversion.  See
!         http://www.netlib.org/lapack/
!***********************************************************************
  SUBROUTINE ls_regr_r(max_n,n,p,x,y,beta_hat)
    USE define_types
!
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: max_n, n, p
    REAL(sp), INTENT(in), DIMENSION(n) :: y
    REAL(sp), INTENT(in), DIMENSION(max_n, p) :: x
    REAL(sp), INTENT(out), DIMENSION(p) :: beta_hat
!
    INTEGER(i4b) :: info
    INTEGER(i4b), DIMENSION(:), ALLOCATABLE :: pivot
    REAL(sp), DIMENSION(:), ALLOCATABLE :: xty
    REAL(sp), DIMENSION(:,:), ALLOCATABLE :: xtx
    REAL(sp), DIMENSION(:), ALLOCATABLE :: work_space
!
    EXTERNAL sgetrf
    EXTERNAL sgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL sgetrf(p,p,xtx,p,pivot,info)
    CALL sgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p) = MATMUL(xtx(1:p,1:p),xty(1:p))
!
!print *, "MATMUL (2) in stat_utilities.f90/ls_regr_r(): OK"
    DEALLOCATE(xtx, xty)
    RETURN
  END SUBROUTINE ls_regr_r
!
  SUBROUTINE ls_regr_d(max_n,n,p,x,y,beta_hat)
    USE define_types
!
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: max_n, n, p
    REAL(dp), INTENT(in), DIMENSION(n) :: y
    REAL(dp), INTENT(in), DIMENSION(max_n, p) :: x
    REAL(dp), INTENT(out), DIMENSION(p) :: beta_hat
!
    INTEGER(i4b) :: info
    INTEGER(i4b), DIMENSION(:), ALLOCATABLE :: pivot
    REAL(dp), DIMENSION(:), ALLOCATABLE :: xty, work_space
    REAL(dp), DIMENSION(:,:), ALLOCATABLE :: xtx

!
    EXTERNAL dgetrf
    EXTERNAL dgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL dgetrf(p,p,xtx,p,pivot,info)
    CALL dgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p) = MATMUL(xtx(1:p,1:p),xty(1:p))
!
    DEALLOCATE(xtx, xty)
    RETURN
  END SUBROUTINE ls_regr_d
!
  SUBROUTINE ls_regr_rmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
    USE define_types
!
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: max_n, n, k, beta_dim, p
    REAL(sp), INTENT(in), DIMENSION(max_n, p) :: x
    REAL(sp), INTENT(in), DIMENSION(max_n, k) :: y
    REAL(sp), INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
!
    INTEGER(i4b) :: info
    INTEGER(i4b), DIMENSION(:), ALLOCATABLE :: pivot
    REAL(sp), DIMENSION(:), ALLOCATABLE :: work_space
    REAL(sp), DIMENSION(:,:), ALLOCATABLE :: xtx, xty
!
    EXTERNAL sgetrf
    EXTERNAL sgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p,k))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n,1:k))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL sgetrf(p,p,xtx,p,pivot,info)
    CALL sgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p,1:k) = MATMUL(xtx(1:p,1:p),xty(1:p,1:k))
!print *, "MATMUL (3) in stat_utilities.f90/ls_regr_rmv(): OK"
!
    DEALLOCATE(xtx, xty)
    RETURN
!
  END SUBROUTINE ls_regr_rmv
!
  SUBROUTINE ls_regr_dmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
    USE define_types
!
    IMPLICIT NONE
    INTEGER(i4b), INTENT(in) :: max_n, n, k, beta_dim, p
    REAL(dp), INTENT(in), DIMENSION(max_n, p) :: x
    REAL(dp), INTENT(in), DIMENSION(max_n, k) :: y
    REAL(dp), INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
!
    INTEGER(i4b) :: info
    INTEGER(i4b), DIMENSION(:), ALLOCATABLE :: pivot
    REAL(dp), DIMENSION(:), ALLOCATABLE :: work_space
    REAL(dp), DIMENSION(:,:), ALLOCATABLE :: xtx, xty
!
    EXTERNAL dgetrf
    EXTERNAL dgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p,k))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n,1:k))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL dgetrf(p,p,xtx,p,pivot,info)
    CALL dgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p,1:k) = MATMUL(xtx,xty)
!print *, "MATMUL (2) in stat_utilities.f90/ls_regr_dmv(): OK"
!
    DEALLOCATE(xtx,xty)
    RETURN
!
  END SUBROUTINE ls_regr_dmv
!
END MODULE stat_utilities
