MODULE ts_utilities
!***********************************************************************
!   FORTRAN 90 module containing commonly used time series
!   statistical computations.
!
!   Created: 03/12/03 JLH
!
!   Contents}:
!        Subroutines for generating vector ARMA realizations:
!            get_mvARMA_params, gen_mvARMA
!        Subroutines for computing optimal AR order:
!             mvaic
!   NOTES: mvaic depends on LAPACK subroutines for computing matrix
!          inverse and LU factorizations for determinant.
!***********************************************************************
  USE define_types
  USE stat_utilities
!
  IMPLICIT NONE
!***********************************************************************
!  Generate mvARMA:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
CONTAINS
  SUBROUTINE get_mvARMA_params(max_k,max_p,max_q,file_label,k,p,q,AR_coefs,MA_coefs)
    USE define_types
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_k, max_p, max_q, file_label, k
    INTEGER(i4b), INTENT(out) :: p, q
    REAL(dp), INTENT(inout), DIMENSION(max_p, max_k, max_k) :: AR_coefs
    REAL(dp), INTENT(inout), DIMENSION(max_q, max_k, max_k) :: MA_coefs
!
    INTEGER(i4b) :: lag, dim1, dim2
!
    WRITE(file_label,400) k
    WRITE(*,*) "Enter the AR order (p) and MA order (q)."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(*,*) "q = ?"
    READ(*,*) q
    WRITE(file_label,500) p, q
!
    IF(p > 0) THEN
       WRITE(*,*) "Enter the AR coefficients, row by row."
       WRITE(file_label,510) 
       DO lag = 1, p
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
    IF(q > 0) THEN
       WRITE(*,*) "Enter the MA coefficients, row by row."
       WRITE(file_label,530) 
       DO lag = 1, q
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
400 FORMAT("Vector autregressive moving average process of dimension ",I1,".")
500 FORMAT("AR order (p) = ",I2,": MA order (q) = ",I2,".")
510 FORMAT("The AR coefficients (row by row) are:")
520 FORMAT("Lag ",I2,":")
530 FORMAT("The MA coefficients (row by row) are:")
!
    RETURN
  END SUBROUTINE get_mvARMA_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvARMA(max_n,max_k,max_p,max_q,n,k,p,q,AR_coefs,MA_coefs,    &
       cov,cov_fac_option,extra,x)
    USE define_types
    USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_n, max_k, max_p, max_q
    INTEGER(i4b), INTENT(in) :: n, k, p, q, cov_fac_option, extra
    REAL(dp), INTENT(in), DIMENSION(max_p, max_k, max_k) :: AR_coefs
    REAL(dp), INTENT(in), DIMENSION(max_q, max_k, max_k) :: MA_coefs
    REAL(dp), INTENT(in), DIMENSION(max_k, k) :: cov
    REAL(dp), INTENT(inout), DIMENSION (max_n, k) :: x
!
    INTEGER(i4b) :: t, j, info
    REAL(dp), DIMENSION(:), ALLOCATABLE :: mean_vec
    REAL(dp), DIMENSION(:,:), ALLOCATABLE :: eps, coefs, work_space
!
    ALLOCATE(eps(max_n,k))
    ALLOCATE(mean_vec(k))
    mean_vec = 0.0_dp
    CALL gen_normal(max_n,max_k,(n+extra),k,mean_vec,cov,cov_fac_option,eps)
    DEALLOCATE(mean_vec)
!
    x = eps
!
    ALLOCATE(work_space(1,k))
    work_space = 0.0_dp
    DO t = p+1,(n+extra)
       DO j = 1, MIN(t-1,p)
          work_space(1,1:k) = MATMUL(x(t-j,1:k),AR_coefs(j,1:k,1:k))
          x(t,1:k) = x(t,1:k) + work_space(1,1:k)
       END DO
       DO j = 1, MIN(t-1,q)
          work_space(1,1:k) = MATMUL(eps(t-j,1:k),MA_coefs(j,1:k,1:k))
          x(t,1:k) = x(t,1:k) - work_space(1,1:k)
       END DO
    END DO
!
    x(1:n,1:k) = x((extra+1):(extra+n),1:k)
    DEALLOCATE(work_space, eps)
!
    RETURN
  END SUBROUTINE gen_mvARMA
!***********************************************************************
!   FORTRAN 90 subroutine for computing AIC and BIC.
!   NOTES: 1.  Uses LAPACK subroutine spotrf (or dpotrf) to compute the 
!          Cholesky decomposition of the covariance matrix.  If LAPACK
!          not available, the Cholesky factorization subroutine in
!          f90_utilities can be used by commenting out the lines
!          EXTERNAL spotrf and IF(cov_fac_option .NE. 0) call spotrf(.)
!          and uncommenting the four lines preceding it.
!
!          2.  Uses LAPACK subroutines dgetrf and dgetri to compute
!          (X^t X)^{-1} in finding least squares estimates for regression
!          coefficients.
!***********************************************************************
  SUBROUTINE mvaic(max_n,n,k,x,max_order,orders,aic,bic)
    USE define_types
    USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER(i4b), INTENT(in) :: max_n, n, k, max_order
    REAL(dp), INTENT(in), DIMENSION(max_n, k) :: x
    INTEGER(i4b), INTENT(out), DIMENSION(max_order) :: orders
    REAL(dp), INTENT(out), DIMENSION(max_order) :: aic, bic
!
    INTEGER(i4b) :: order, no_obs, col_x, last_col_x
    INTEGER(i4b) :: t, dim, lag, index, info, ii, jj
    INTEGER(i4b), DIMENSION(:), ALLOCATABLE :: pivot
    REAL(dp) :: dk, dno_obs, det_cov, d, k2o
    REAL(dp), DIMENSION(:,:), ALLOCATABLE :: work_space, beta_hat
    REAL(dp), DIMENSION(:,:), ALLOCATABLE :: fitted, residuals, cov_hat
!
    EXTERNAL dgetrf
!
    dk = DBLE(k)
    aic = 0.0_dp
    bic = 0.0_dp
    orders = 0
!
    ALLOCATE(cov_hat(k,k))
    DO order = 1, max_order
       no_obs = n - order
       col_x = k*order
       last_col_x = k + col_x
       orders(order) = order
!
       ALLOCATE(work_space(no_obs,last_col_x))
       DO t = 1, no_obs
          work_space(t,1:k) = x(t+order,1:k)
          DO dim = 1, k
             DO lag = 1, order
                index = k*lag + dim
                work_space(t,index) = x(t+lag-1,dim)
             END DO
          END DO
       END DO
!
       ALLOCATE(beta_hat(col_x,k))
       ALLOCATE(fitted(no_obs,k))
       ALLOCATE(residuals(no_obs,k))
!
       CALL ls_regression(max_n,n,k,col_x,work_space(1:no_obs,(k+1):last_col_x),  &
	        work_space(1:no_obs,1:k),col_x,beta_hat)
!
       fitted = MATMUL(work_space(1:no_obs,(k+1):last_col_x),beta_hat)
       residuals = work_space(1:no_obs,1:k) - fitted(1:no_obs,1:k)
       cov_hat = MATMUL(TRANSPOSE(residuals),residuals)
!
       ALLOCATE(pivot(k))
       CALL dgetrf(k,k,cov_hat,k,pivot,info)       
       DEALLOCATE(pivot)
!
       det_cov = 1.0_dp
       DO dim = 1,k
          det_cov = det_cov*ABS(cov_hat(dim,dim))
       END DO
!
       DEALLOCATE(work_space, beta_hat, fitted, residuals)
!
       dno_obs = DBLE(no_obs)
       d = dno_obs * LOG(det_cov) 
       k2o = dk * dk * DBLE(order)
!
       aic(order) = d + 2.0_dp*k2o
       bic(order) = d + LOG(dno_obs)*k2o
    END DO
    DEALLOCATE(cov_hat)
!
    RETURN
  END SUBROUTINE mvaic
!
END MODULE ts_utilities
